from sqlalchemy import Select, inspect, select
from sqlalchemy.orm import load_only, selectinload

from fastgear.types.find_many_options import FindManyOptions
from fastgear.types.find_one_options import FindOneOptions
from fastgear.types.generic_types_var import EntityType


class SelectConstructor:
    def __init__(self, entity: EntityType) -> None:
        self.entity = entity

    def build_select_statement(
        self, criteria: str | FindOneOptions | FindManyOptions = None, new_entity: EntityType = None,
    ) -> Select:
        """Constructs and returns a SQLAlchemy Select statement based on the provided criteria and entity.

        Args:
            criteria (str | FindOneOptions | FindManyOptions, optional): The filter criteria to build the select
                statement. It can be a string, an instance of FindOneOptions, or an instance of FindManyOptions.
                Defaults to None.
            new_entity (EntityType, optional): A new entity type to use for the select statement.
                If not provided, the existing entity type will be used. Defaults to None.

        Returns:
            Select: The constructed SQLAlchemy Select statement.

        """
        entity = new_entity or self.entity

        if isinstance(criteria, str):
            criteria = self.__generate_find_one_options_dict(criteria, entity)

        select_statement = select(entity)

        return self.__apply_options(select_statement, entity, criteria)

    def __apply_options(
        self,
        select_statement: Select,
        entity: EntityType,
        options_dict: FindOneOptions | FindManyOptions = None,
    ) -> Select:
        """Applies various options to the given SQLAlchemy Select statement based on the provided option's dictionary.

        Args:
            select_statement (Select): The initial SQLAlchemy Select statement to which options will be applied.
            entity (EntityType): The entity type associated with the select statement.
            options_dict (FindOneOptions | FindManyOptions, optional): A dictionary containing various options to be
                applied to the select statement. Defaults to None.

        Returns:
            Select: The modified SQLAlchemy Select statement with the applied options.

        """
        if not options_dict:
            return select_statement

        options_dict = self.__fix_options_dict(options_dict)

        for key in options_dict:
            match key:
                case "select":
                    select_statement = select_statement.options(
                        load_only(*options_dict[key], raiseload=True),
                    )
                case "where":
                    select_statement = select_statement.where(*options_dict[key])
                case "order_by":
                    select_statement = select_statement.order_by(*options_dict[key])
                case "skip":
                    select_statement = select_statement.offset(options_dict[key])
                case "take":
                    select_statement = select_statement.limit(options_dict[key])
                case "relations":
                    select_statement = select_statement.options(
                        *[selectinload(getattr(entity, relation)) for relation in options_dict[key]],
                    )
                case _:
                    raise KeyError(f"Unknown option: {key} in FindOptions")

        return select_statement

    @staticmethod
    def extract_from_mapping(field_mapping: dict, fields: list) -> list:
        """Extracts and returns a list of items from the field mapping based on the provided fields.

        Args:
            field_mapping (dict): A dictionary mapping fields to their corresponding items.
            fields (list): A list of fields to extract items for.

        Returns:
            list: A list of items extracted from the field mapping based on the provided fields.

        """
        return [
            item
            for field in fields
            for item in (
                field_mapping.get(field, [field])
                if isinstance(field_mapping.get(field, field), list)
                else [field_mapping.get(field, field)]
            )
        ]

    @staticmethod
    def __fix_options_dict(
        options_dict: FindOneOptions | FindManyOptions,
    ) -> FindOneOptions | FindManyOptions:
        """Ensures that specific attributes in the options dictionary are lists.

        Args:
            options_dict (FindOneOptions | FindManyOptions): The options dictionary to be fixed.

        Returns:
            FindOneOptions | FindManyOptions: The fixed options dictionary with specific attributes as lists.

        """
        for attribute in ["where", "order_by", "options"]:
            if attribute in options_dict and not isinstance(options_dict[attribute], list):
                options_dict[attribute] = [options_dict[attribute]]

        return options_dict

    @staticmethod
    def __generate_find_one_options_dict(criteria: str, entity: EntityType) -> FindOneOptions:
        """Generates a FindOneOptions dictionary based on the provided criteria and entity.

        Args:
            criteria (str): The criteria to filter the entity. Typically, this is the primary key value.
            entity (EntityType): The entity type for which the options dictionary is being generated.

        Returns:
            FindOneOptions: A dictionary with a 'where' clause that filters the entity based on the primary key.

        """
        return {"where": [inspect(entity).primary_key[0] == criteria]}



import logging
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, Generic, TypeVar

from pydantic import BaseModel

from fastgear.common.database.sqlalchemy.repository_utils.base_repository_utils import (
    BaseRepositoryUtils,
)
from fastgear.common.database.sqlalchemy.repository_utils.select_constructor import (
    SelectConstructor,
)
from fastgear.types.delete_result import DeleteResult
from fastgear.types.find_many_options import FindManyOptions
from fastgear.types.find_one_options import FindOneOptions
from fastgear.types.update_result import UpdateResult

EntityType = TypeVar("EntityType")
SessionType = TypeVar("SessionType")


class AbstractRepository(ABC, Generic[EntityType]):
    def __init__(self, entity: type[EntityType]) -> None:
        self.entity = entity
        self.select_constructor = SelectConstructor(entity)
        self.repo_utils = BaseRepositoryUtils()

        self.logger = logging.getLogger(self.__class__.__module__)

    @abstractmethod
    def create(self, new_record: EntityType | Any, db: SessionType) -> EntityType:
        """Abstract method to create a new record.

        Args:
            new_record (EntityType | Any): The new record to be created. It can be an instance of
                EntityType or any other type.
            db (SessionType): The database session.

        Returns:
            EntityType: The created record.
        """

    @abstractmethod
    def create_all(
        self, new_records: list[EntityType | Any], db: SessionType = None
    ) -> list[EntityType]:
        """Abstract method to create multiple records.

        Args:
            new_records (List[EntityType | Any]): A list of new records to be created. Each record
                can be an instance of EntityType or any other type.
            db (SessionType): The database session.

        Returns:
            List[EntityType]: A list of the created records.
        """

    @staticmethod
    @abstractmethod
    def save(
        new_record: EntityType | list[EntityType] = None, db: SessionType = None
    ) -> EntityType | list[EntityType] | None:
        """Abstract method to save records.

        Args:
            new_record (EntityType | List[EntityType], optional): The new record or list of records
                to be saved. Defaults to None.
            db (SessionType): The database session.

        Returns:
            EntityType | List[EntityType] | None: The saved record or list of records, or None
                if no records were saved.
        """

    @staticmethod
    @abstractmethod
    def commit_or_flush(db: SessionType = None) -> None:
        """Abstract method to commit or flush.

        Args:
            db (SessionType): The database session.

        Returns:
            None
        """

    @staticmethod
    @abstractmethod
    def refresh_record(
        new_record: EntityType | list[EntityType], db: SessionType = None
    ) -> EntityType | list[EntityType]:
        """Abstract method to refresh records.

        Args:
            new_record (EntityType | List[EntityType]): The record or list of records
                to be refreshed.
            db (SessionType): The database session.

        Returns:
            EntityType | List[EntityType]: The refreshed record or list of records.
        """

    @abstractmethod
    def find_one(
        self, search_filter: str | FindOneOptions, db: SessionType = None
    ) -> EntityType | None:
        """Abstract method to find one record.

        Args:
            search_filter (str | FindOneOptions): The filter criteria to find the record.
                It can be a string or an instance of FindOneOptions.
            db (SessionType): The database session.

        Returns:
            EntityType | None: The found record or None if no record matches the filter criteria.
        """

    @abstractmethod
    def find_one_or_fail(
        self, search_filter: str | FindOneOptions, db: SessionType = None
    ) -> EntityType:
        """Abstract method to find one record or fail if no record is found.

        Args:
            search_filter (str | FindOneOptions): The filter criteria to find the record.
                It can be a string or an instance of FindOneOptions.
            db (SessionType): The database session.

        Returns:
            EntityType: The found record.

        Raises:
            NotFoundException: If no record matches the filter criteria.
        """

    @abstractmethod
    def find(
        self, stmt_or_filter: FindManyOptions | Any = None, db: SessionType = None
    ) -> Sequence[EntityType]:
        """Abstract method to find many records.

        Args:
            stmt_or_filter (FindManyOptions | Any, optional): The filter criteria or statement to
                find the records. Defaults to None.
            db (SessionType): The database session.

        Returns:
            Sequence[EntityType]: A sequence of the found records.
        """

    @abstractmethod
    def count(self, stmt_or_filter: FindManyOptions | Any = None, db: SessionType = None) -> int:
        """Abstract method to count without limit and offset the number of rows returned by a
            Select Statement. Implementation can be sync or async.

        Args:
            stmt_or_filter (FindManyOptions | Any, optional): The filter criteria or statement to
                count the records. Defaults to None.
            db (SessionType): The database session.

        Returns:
            int: The count of the records.
        """

    @abstractmethod
    def find_and_count(
        self, search_filter: FindManyOptions = None, db: SessionType = None
    ) -> tuple[Sequence[EntityType], int]:
        """Abstract method to find and count records. Implementation can be sync or async.

        Args:
            search_filter (FindManyOptions, optional): The filter criteria to find the records.
                Defaults to None.
            db (SessionType): The database session.

        Returns:
            Tuple[Sequence[EntityType], int]: A tuple containing a sequence of the found records
                and the count of the records.
        """

    @abstractmethod
    def update(
        self,
        search_filter: str | FindOneOptions,
        model_data: BaseModel | dict,
        db: SessionType = None,
    ) -> UpdateResult:
        """Abstract method to update records. Implementation can be sync or async.

        Args:
            search_filter (str | FindOneOptions): The filter criteria to find the record to be
                updated. It can be a string or an instance of FindOneOptions.
            model_data (BaseModel | dict): The data to update the record with. It can be an
                instance of BaseModel or a dictionary.
            db (SessionType): The database session.

        Returns:
            UpdateResult: The result of the update operation.
        """

    @abstractmethod
    def delete(
        self, delete_statement: str | FindOneOptions | Any, db: SessionType = None
    ) -> DeleteResult:
        """Abstract method to delete records. Implementation can be sync or async.

        Args:
            delete_statement (str | FindOneOptions | Any): The statement or filter criteria to
                delete the records. It can be a string, an instance of FindOneOptions, or any
                other type.
            db (SessionType): The database session.

        Returns:
            DeleteResult: The result of the delete operation.
        """



from .enum_utils import EnumUtils
from .json_utils import JsonUtils
from .logger_utils import LoggerUtils
from .pagination_utils import PaginationUtils

__all__ = ["EnumUtils", "JsonUtils", "LoggerUtils", "PaginationUtils"]



import asyncio
from collections.abc import Callable
from functools import singledispatchmethod, wraps
from inspect import iscoroutinefunction, signature
from types import UnionType
from typing import Any, TypeVar, Union, get_args, get_origin

from fastgear.common.database.sqlalchemy.session import AllSessionType, db_session

ClassType = TypeVar("ClassType")


def inject_db_parameter_decorator(cls: type[ClassType]) -> type[ClassType]:
    """Class decorator that modifies methods of the given class to automatically inject a default parameter value
    if it is not already present in the method's arguments. It applies another decorator,
    `inject_db_parameter_if_missing`, to each method.

    Args:
        cls (Type[ClassType]): The class to be decorated.

    Returns:
        Type[ClassType]: The decorated class with methods that automatically inject a default parameter value.

    """

    def decorate_method(attr_value: Any) -> Any:
        if isinstance(attr_value, staticmethod | classmethod):
            func = attr_value.__func__
            decorated_func = inject_db_parameter_if_missing(func)
            return type(attr_value)(decorated_func)

        if isinstance(attr_value, singledispatchmethod):
            dispatcher = attr_value.dispatcher
            # Wrap all registered implementations
            for typ, func in dispatcher.registry.items():
                # Wrap the function
                wrapped_func = inject_db_parameter_if_missing(func)
                # Re-register the function
                dispatcher.register(typ, wrapped_func)
            # Reconstruct the singledispatchmethod
            return singledispatchmethod(dispatcher)

        if callable(attr_value):
            # Regular method
            return inject_db_parameter_if_missing(attr_value)

        return attr_value

    # Iterate over all class attributes to find methods (excluding magic methods)
    for attr_name, attr_value in cls.__dict__.items():
        if attr_name.startswith("__"):
            continue
        new_attr = decorate_method(attr_value)
        setattr(cls, attr_name, new_attr)

    return cls


def inject_db_parameter_if_missing(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator that injects a Session or AsyncSession instance into the function arguments
    if it is missing and required by the function's signature.
    """

    def db_session_injection(*args: tuple[Any], **kwargs: Any) -> (tuple[Any], Any):
        sig = signature(func)
        params = list(sig.parameters.values())

        # Check if a Session instance is not among the arguments and if injection is needed
        if not any(isinstance(arg, AllSessionType) for arg in args):
            for i, param in enumerate(params):
                if (
                    param.default is None
                    and param.name not in kwargs
                    and len(args) <= i
                    and is_valid_session_type(param.annotation)
                ):
                    kwargs[param.name] = db_session.get()

        return args, kwargs

    def is_valid_session_type(annotation: Any) -> bool:
        origin = get_origin(annotation)
        if origin in (Union, UnionType):
            return any(
                isinstance(cls, type) and issubclass(cls, AllSessionType)
                for cls in get_args(annotation)
            )
        return isinstance(annotation, type) and issubclass(annotation, AllSessionType)

    @wraps(func)
    async def wrapper(*args, **kwargs) -> Any:
        args, kwargs = db_session_injection(*args, **kwargs)
        if iscoroutinefunction(func):
            return await func(*args, **kwargs)
        await asyncio.to_thread(func, *args, **kwargs)
        return None

    return wrapper



from pydantic import BaseModel

from fastgear.types.generic_types_var import EntityType


class BaseRepositoryUtils:
    @staticmethod
    def should_be_updated(entity: EntityType, update_schema: BaseModel) -> bool:
        """Determines if the given entity should be updated based on the provided update schema.

        Args:
            entity (EntityType): The entity to check for updates.
            update_schema (BaseModel): The schema containing the update data.

        Returns:
            bool: True if the entity should be updated, False otherwise.

        """
        return any(
            getattr(entity, key) != value
            for key, value in update_schema.model_dump(exclude_unset=True).items()
        )



import logging
import sys


class LoggerUtils:
    @staticmethod
    def configure_logging(level: int | str) -> None:
        logger = logging.getLogger()
        if not logger.handlers:
            handler = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter("%(asctime)s - %(name)s - [%(levelname)s]: %(message)s")
            handler.setFormatter(formatter)

            logger.setLevel(level)
            logger.addHandler(handler)



from datetime import datetime
from typing import Any
from uuid import UUID

from pydantic import BaseModel
from sqlalchemy.exc import InvalidRequestError

from fastgear.common.schema.custom_base_model_schema import CustomBaseModel


class BaseSchema(CustomBaseModel):
    """Base schema class that includes common fields and a method to validate and exclude unloaded fields.

    Attributes:
        id (UUID): Unique identifier for the schema.
        created_at (datetime | None): Timestamp when the record was created.
        updated_at (datetime | None): Timestamp when the record was last updated.

    """

    id: UUID
    created_at: datetime | None = None
    updated_at: datetime | None = None

    @classmethod
    def model_validate_exclude_unloaded(cls: type[BaseModel], obj: Any) -> BaseModel:
        """Validates the model and excludes fields that are not loaded.

        This method attempts to get the attribute for each field in the model. If an InvalidRequestError
        is raised, it means the field is not loaded, and it is excluded from the validation.

        Args:
            cls (Type[BaseModel]): The class type of the model.
            obj (Any): The object to validate, can be a dictionary or an instance of the model.

        Returns:
            BaseModel: An instance of the model with only the loaded fields.

        """
        obj_dict = {}
        if isinstance(obj, dict):
            obj_dict = obj
        else:
            for field_name in cls.__fields__:
                try:
                    # Attempt to get the attribute, if it raises an InvalidRequestError, it is not loaded
                    value = getattr(obj, field_name)
                    obj_dict[field_name] = value
                except InvalidRequestError:
                    # Don't add the field if it is not loaded
                    continue

        return cls(**obj_dict)



import redis.asyncio as aioredis
from redis.asyncio import Redis


class RedisClientFactory:
    """Factory class for creating and managing Redis client connections using an asynchronous connection pool.

    Args:
        redis_url (str): The URL of the Redis server.

    """

    def __init__(self, redis_url: str) -> None:
        self.pool = aioredis.ConnectionPool.from_url(redis_url)

    def get_client(self) -> Redis:
        """Retrieves a Redis client instance from the connection pool.

        Returns:
            Redis: An instance of the Redis client.

        """
        return aioredis.Redis(connection_pool=self.pool)

    async def close_pool(self) -> None:
        """Closes the connection pool asynchronously.

        Returns:
            None

        """
        await self.pool.aclose()



from typing_extensions import TypedDict


class PaginationSearch(TypedDict):
    field: str
    value: str


class PaginationSort(TypedDict):
    field: str
    by: str


class Pagination(TypedDict):
    skip: int
    take: int
    sort: list[PaginationSort]
    search: list[PaginationSearch]
